"""Manage sample data."""

import os
import subprocess
from importlib.resources import files

__all__ = [
    "get_sample_path",
    "make_all_samples",
    "clean_all_samples",
]


def get_sample_path(*paths, build=True):
    """Return path to sample data distributed by the package.

    If file does not exist, attempts to generate it using pre-defined recipe.

    Parameters
    ----------
    paths : list of str
        Subpath components.
    build : bool, default=True
        Attempt to build if file is missing.

    Returns
    -------
    pathlib.Path

    Examples
    --------
    >>> from heavyedge_features.samples import get_sample_path
    >>> print(get_sample_path())  #doctest: +SKIP
    heavyedge_features/samples
    >>> print(get_sample_path("RawData"))  #doctest: +SKIP
    heavyedge_features/samples/RawData
    """
    datapath = files("heavyedge_features.samples")
    path = datapath.joinpath(*paths)

    if not os.path.exists(path) and build:
        from .recipes import RECIPES

        recipe = RECIPES.get("/".join(paths))
        if recipe is not None:
            commands = recipe(path)
            if commands is not None:
                subprocess.run(commands, check=True)
    return path


def make_all_samples(progress=False):
    """Generate all sample data.

    Parameters
    ----------
    progress : bool, default=False
        Prints progress bar. Requires :mod:`tqdm`.
    """
    from .recipes import RECIPES

    if progress:
        from tqdm import tqdm

        items = tqdm(RECIPES.items())
    else:
        items = RECIPES.items()

    for file, recipe in items:
        path = get_sample_path(file, build=False)
        if not os.path.exists(path):
            commands = recipe(path)
            if commands is not None:
                subprocess.run(commands, check=True)


def clean_all_samples():
    """Delete all buildable samples."""
    from .recipes import RECIPES

    for file in RECIPES.keys():
        path = get_sample_path(file, build=False)
        path.unlink(missing_ok=True)



"""Recipies to build sample data."""

import numpy as np
from heavyedge import ProfileData
from heavyedge import get_sample_path as heavyedge_sample
from heavyedge_classify.samples import get_sample_path as classify_sample

from . import get_sample_path


def save_wetthicknesses(path):
    N, _ = ProfileData(get_sample_path("Profiles.h5")).shape()
    wet_thicknesses = np.full(N, 0.25)
    np.save(path, wet_thicknesses)


RECIPES = {
    "Profiles.h5": lambda path: [
        "heavyedge",
        "merge",
        heavyedge_sample("Prep-Type1.h5"),
        heavyedge_sample("Prep-Type2.h5"),
        heavyedge_sample("Prep-Type3.h5"),
        "-o",
        path,
    ],
    "global-features.csv": lambda path: [
        "heavyedge",
        "features-global",
        classify_sample("labels-pred.npy"),
        "--target-indices",
        "0",
        "-o",
        path,
    ],
    "wet_thickness.npy": lambda path: save_wetthicknesses(path),
    "local-features.csv": lambda path: [
        "heavyedge",
        "features-local",
        get_sample_path("Profiles.h5"),
        classify_sample("labels-pred.npy"),
        get_sample_path("wet_thickness.npy"),
        "--sigma",
        "32",
        "--type1-indices",
        "0",
        "--type2-indices",
        "1",
        "--type3-indices",
        "2",
        "-o",
        path,
    ],
}



"""High-level Python runtime interface."""

import numpy as np

from ..edge_width import width_type0, width_type1, width_type2, width_type3
from ..iproj import signed_iproj

__all__ = [
    "global_deviation",
    "edge_height",
    "edge_width",
]


def global_deviation(
    soft_labels,
    target_indices,
    logger=lambda x: None,
):
    """Compute global shape deviations using probabilistic classification labels.

    Negative values indicaete profiles are within the desired classes.
    Larger values mean more deviation from the desired classes.

    Parameters
    ----------
    soft_labels : np.ndarray
        Probabilistic classification labels for the profiles.
    target_indices : list of int
        Indices of target classes to compute values for.
    logger : callable, optional
        Logger function which accepts a progress message string.

    Returns
    -------
    values : np.ndarray
        Array containing global shape deviations for each profile.

    Examples
    --------
    >>> import numpy as np
    >>> from heavyedge_classify.samples import get_sample_path
    >>> from heavyedge_features.api import global_deviation
    >>> soft_labels = np.load(get_sample_path("labels-pred.npy"))
    >>> global_deviation(soft_labels, [0]).shape
    (75,)
    """
    N, _ = soft_labels.shape
    values = []
    for i, p in enumerate(soft_labels):
        value, _ = signed_iproj(p, target_indices)
        values.append(value)
        logger(f"{i + 1}/{N}")
    return np.array(values)


def edge_height(profiles, logger=lambda x: None):
    """Dimensionless edge height of edge profiles.

    Parameters
    ----------
    profiles : heavyedge.ProfileData
        Open h5 file of profiles.
    logger : callable, optional
        Logger function which accepts a progress message string.

    Returns
    -------
    heights : np.ndarray
        Array containing edge height values for each profile.

    Examples
    --------
    >>> from heavyedge import ProfileData
    >>> from heavyedge_features.samples import get_sample_path as features_sample
    >>> from heavyedge_features.api import edge_height
    >>> edge_height(ProfileData(features_sample("Profiles.h5"))).shape
    (75,)
    """
    N, _ = profiles.shape()
    ret = []
    for i, (Y, L, _) in enumerate(profiles):
        ret.append(Y[:L].max() / Y[0])
        logger(f"edge height ({i + 1}/{N})")
    return np.array(ret)


def edge_width(
    profiles,
    hard_labels,
    wet_thicknesses,
    sigma,
    type1_indices,
    type2_indices,
    type3_indices,
    logger=lambda x: None,
):
    """Detect edge with of profiles using profile data and classification labels.

    Parameters
    ----------
    profiles : heavyedge.ProfileData
        Open h5 file of profiles.
    hard_labels : np.ndarray
        Hard classification labels for the profiles.
    wet_thicknesses : np.ndarray
        Wet thickness values for the profiles.
    sigma : scalar
        Standard deviation of Gaussian filter for smoothing.
        Using the same value as the one used for preprocessing is recommended.
    type1_indices, type2_indices, type3_indices : list of int
        Lists of indices of Type 1, 2, and 3 classes from trained labels, respectively.
    logger : callable, optional
        Logger function which accepts a progress message string.

    Returns
    -------
    widths : np.ndarray
        Array containing edge width values for each profile.

    Examples
    --------
    >>> from heavyedge import ProfileData
    >>> from heavyedge_classify.samples import get_sample_path as classify_sample
    >>> from heavyedge_features.samples import get_sample_path as features_sample
    >>> from heavyedge_features.api import edge_width
    >>> import numpy as np
    >>> profiles = ProfileData(features_sample("Profiles.h5"))
    >>> hard_labels = np.load(classify_sample("labels-pred.npy")).argmax(axis=1)
    >>> wet_thicknesses = np.full(hard_labels.shape, 0.25)
    >>> sigma = 32
    >>> edge_width(profiles, hard_labels, wet_thicknesses, sigma, [0], [1], [2]).shape
    (75,)
    """
    x = profiles.x()
    N, _ = profiles.shape()
    ret = []
    for i, ((Y, L, _), label, wt) in enumerate(
        zip(profiles, hard_labels, wet_thicknesses)
    ):
        if label in type1_indices:
            width = width_type1(x, Y, L, wt)
        elif label in type2_indices:
            width = width_type2(x, Y, L, sigma)
        elif label in type3_indices:
            width = width_type3(x, Y, L, sigma)
        else:
            width = width_type0(x, Y, L, wt)
        logger(f"edge width ({i + 1}/{N})")
        ret.append(width)
    return np.array(ret)



import os
import subprocess


def test_globalfeatures_label_csv(tmp_data_path, tmp_path):
    _, (_, label_csv_path), _ = tmp_data_path
    out_path = tmp_path / "global-features.csv"
    subprocess.run(
        [
            "heavyedge",
            "--log-level=INFO",
            "features-global",
            label_csv_path,
            "--target-indices",
            "0",
            "-o",
            out_path,
        ],
        check=True,
    )
    assert os.path.exists(out_path)


def test_localfeatures_label_csv(tmp_data_path, tmp_path):
    profiles_path, (_, label_csv_path), wt_path = tmp_data_path
    out_path = tmp_path / "local-features.csv"
    subprocess.run(
        [
            "heavyedge",
            "--log-level=INFO",
            "features-local",
            profiles_path,
            label_csv_path,
            wt_path,
            "--sigma",
            "1.0",
            "--type1-indices",
            "0",
            "--type2-indices",
            "1",
            "--type3-indices",
            "2",
            "-o",
            out_path,
        ],
        check=True,
    )



"""Edge width by edge type."""

import numpy as np
from heavyedge_landmarks import (
    landmarks_type2,
    landmarks_type3,
    plateau_type2,
    plateau_type3,
)

__all__ = [
    "width_type0",
    "width_type1",
    "width_type2",
    "width_type3",
]


def width_type0(x, Y, L, wt):
    """Edge width for type 0 profiles.

    Parameters
    ----------
    x : array of shape (M,)
        X grid of profiles.
    Y : array of shape (M,)
        Height data of profile.
    L : int
        Length of profile before contact point in number of points.
    wt : scalar
        Wet thickness of the profile.

    Returns
    -------
    scalar
        Edge width of the profile.
    """
    (idxs,) = np.where(Y <= wt)
    if len(idxs) == 0:
        edge_idx = 0
    else:
        edge_idx = idxs[0]
    return x[L - 1] - x[edge_idx]


def width_type1(x, Y, L, wt):
    """Edge width for type 1 profiles.

    Parameters
    ----------
    x : array of shape (M,)
        X grid of profiles.
    Y : array of shape (M,)
        Height data of profile.
    L : int
        Length of profile before contact point in number of points.
    wt : scalar
        Wet thickness of the profile.

    Returns
    -------
    scalar
        Edge width of the profile.

    Examples
    --------
    >>> from heavyedge import get_sample_path, ProfileData
    >>> from heavyedge_features.edge_width import width_type1
    >>> with ProfileData(get_sample_path("Prep-Type1.h5")) as data:
    ...     x = data.x()
    ...     Y, L, _ = data[0]
    >>> wt = 0.25
    >>> b = width_type1(x, Y, L, wt)
    >>> import matplotlib.pyplot as plt  # doctest: +SKIP
    ... plt.plot(x, Y, color="gray", alpha=0.5)
    ... plt.axhline(wt, ls="--", label="Wet thickness")
    ... plt.axvline(x[L - 1] - b, color="red", ls="--", label="Edge boundary")
    ... plt.legend()
    """
    (idxs,) = np.where(Y >= wt)
    if len(idxs) == 0:
        edge_idx = 0
    else:
        edge_idx = idxs[-1]
    return x[L - 1] - x[edge_idx]


def width_type2(x, Y, L, sigma):
    """Edge width for type 2 profiles.

    Parameters
    ----------
    x : array of shape (M,)
        X grid of profiles.
    Y : array of shape (M,)
        Height data of profile.
    L : int
        Length of profile before contact point in number of points.
    sigma : scalar
        Standard deviation of Gaussian filter for smoothing.

    Returns
    -------
    scalar
        Edge width of the profile.

    Examples
    --------
    >>> from heavyedge import get_sample_path, ProfileData
    >>> from heavyedge_features.edge_width import width_type2
    >>> with ProfileData(get_sample_path("Prep-Type2.h5")) as data:
    ...     x = data.x()
    ...     Y, L, _ = data[0]
    >>> sigma = 32.0
    >>> b = width_type2(x, Y, L, sigma)
    >>> import matplotlib.pyplot as plt  # doctest: +SKIP
    ... plt.plot(x, Y, color="gray", alpha=0.5)
    ... plt.axvline(x[L - 1] - b, color="red", ls="--", label="Edge boundary")
    ... plt.legend()
    """
    lm = landmarks_type2(x, [Y], [L], sigma)
    peaks, knees = lm[:, 0, 1:].T
    plateau = plateau_type2(x, [Y], peaks, knees)
    return x[L - 1] - plateau[0, -1]


def width_type3(x, Y, L, sigma):
    """Edge width for type 3 profiles.

    Parameters
    ----------
    x : array of shape (M,)
        X grid of profiles.
    Y : array of shape (M,)
        Height data of profile.
    L : int
        Length of profile before contact point in number of points.
    sigma : scalar
        Standard deviation of Gaussian filter for smoothing.

    Returns
    -------
    scalar
        Edge width of the profile.

    Examples
    --------
    >>> from heavyedge import get_sample_path, ProfileData
    >>> from heavyedge_features.edge_width import width_type3
    >>> with ProfileData(get_sample_path("Prep-Type3.h5")) as data:
    ...     x = data.x()
    ...     Y, L, _ = data[0]
    >>> sigma = 32.0
    >>> b = width_type3(x, Y, L, sigma)
    >>> import matplotlib.pyplot as plt  # doctest: +SKIP
    ... plt.plot(x, Y, color="gray", alpha=0.5)
    ... plt.axvline(x[L - 1] - b, color="red", ls="--", label="Edge boundary")
    ... plt.legend()
    """
    lm = landmarks_type3(x, [Y], [L], sigma)
    troughs, knees = lm[:, 0, 2:].T
    plateau = plateau_type3(x, [Y], troughs, knees)
    return x[L - 1] - plateau[0, -1]



"""Commands for edge feature quantification."""

import pathlib

from heavyedge.cli.command import Command, register_command

PLUGIN_ORDER = 2.0


@register_command("features-global", "Quantify global shape features")
class GlobalFeaturesCommand(Command):
    def add_parser(self, main_parser):
        parser = main_parser.add_parser(
            self.name,
            help="Quantify global shape feature score using classification model",
            epilog=(
                "The input label can be in npy (default) or csv format. "
                "If csv, the first row is the header. "
                "Unrecognized formats are parsed as npy with a warning. "
                "Output field 'phi' is the signed distance to the "
                "nearest admissible class in the probability simplex."
            ),
        )
        parser.add_argument(
            "soft_labels",
            type=pathlib.Path,
            help="Path to file containing probabilistic classification labels.",
        )
        parser.add_config_argument(
            "--target-indices",
            type=int,
            nargs="+",
            help="List of indices of admissible classes from trained labels.",
        )
        parser.add_argument(
            "--label-format",
            choices=["npy", "csv"],
            help="Label file format. If not passed, parsed from file extension.",
        )
        parser.add_argument(
            "-o",
            "--output",
            type=pathlib.Path,
            help="Path to output csv file",
        )

    def run(self, args):
        import csv
        import os

        import numpy as np

        from heavyedge_features.api import global_deviation

        if args.target_indices is None:
            raise ValueError("--target-indices must be specified.")

        label_ext = os.path.splitext(args.soft_labels)[1].lower().lstrip(".")
        label_format = args.label_format or label_ext

        self.logger.info(f"Start processing {args.output}")

        if label_format == "csv":
            with open(args.soft_labels, "r") as f:
                reader = csv.reader(f)
                # Burn first row as header
                next(reader)
                soft_labels = np.array([row for row in reader], dtype=float)
        else:
            if label_format != "npy":
                self.logger.warning(
                    f"Unrecognized label format '{label_format}', parsing as npy."
                )
            soft_labels = np.load(args.soft_labels)

        values = global_deviation(
            soft_labels,
            args.target_indices,
            logger=lambda msg: self.logger.info(f"{args.output} : {msg}"),
        )

        with open(args.output, "w", newline="") as f:
            writer = csv.writer(f)
            writer.writerow(["phi"])
            for value in values:
                writer.writerow([value])

        self.logger.info(f"Saved {args.output}.")


@register_command("features-local", "Quantify local shape features")
class LocalFeaturesCommand(Command):
    def add_parser(self, main_parser):
        parser = main_parser.add_parser(
            self.name,
            help="Quantify local shape features",
            epilog=(
                "The input label can be in npy (default) or csv format. "
                "If csv, the first row is the header. "
                "Unrecognized formats are parsed as npy with a warning. "
                "Output field 'H' is the apparent edge superelevation. "
                "Output field 'b' is the edge width."
            ),
        )
        parser.add_argument(
            "profiles",
            type=pathlib.Path,
            help="h5 file path to profile data in 'ProfileData' structure.",
        )
        parser.add_argument(
            "soft_labels",
            type=pathlib.Path,
            help="Path to file containing probabilistic classification labels.",
        )
        parser.add_argument(
            "h_w",
            type=pathlib.Path,
            help="Path to npy file containing wet thickness.",
        )
        parser.add_config_argument(
            "--sigma",
            type=float,
            help="Standard deviation of Gaussian kernel.",
        )
        parser.add_config_argument(
            "--type1-indices",
            type=int,
            nargs="+",
            help="List of indices of Type 1 classes from trained labels.",
        )
        parser.add_config_argument(
            "--type2-indices",
            type=int,
            nargs="+",
            help="List of indices of Type 2 classes from trained labels.",
        )
        parser.add_config_argument(
            "--type3-indices",
            type=int,
            nargs="+",
            help="List of indices of Type 3 classes from trained labels.",
        )
        parser.add_argument(
            "--label-format",
            choices=["npy", "csv"],
            help="Label file format. If not passed, parsed from file extension.",
        )
        parser.add_argument(
            "-o",
            "--output",
            type=pathlib.Path,
            help="Path to output csv file",
        )

    def run(self, args):
        import csv
        import os

        import numpy as np
        from heavyedge import ProfileData

        from heavyedge_features.api import edge_height, edge_width

        if args.type1_indices is None:
            raise ValueError("--type1-indices must be specified.")
        if args.type2_indices is None:
            raise ValueError("--type2-indices must be specified.")
        if args.type3_indices is None:
            raise ValueError("--type3-indices must be specified.")

        label_ext = os.path.splitext(args.soft_labels)[1].lower().lstrip(".")
        label_format = args.label_format or label_ext

        self.logger.info(f"Start processing {args.output}")

        if label_format == "csv":
            with open(args.soft_labels, "r") as f:
                reader = csv.reader(f)
                # Burn first row as header
                next(reader)
                soft_labels = np.array([row for row in reader], dtype=float)
        else:
            if label_format != "npy":
                self.logger.warning(
                    f"Unrecognized label format '{label_format}', parsing as npy."
                )
            soft_labels = np.load(args.soft_labels)

        wet_thicknesses = np.load(args.h_w)

        edge_heights = edge_height(
            ProfileData(args.profiles),
            logger=lambda msg: self.logger.info(f"{args.output} : {msg}"),
        )
        edge_widths = edge_width(
            ProfileData(args.profiles),
            soft_labels.argmax(axis=1),
            wet_thicknesses,
            args.sigma,
            args.type1_indices,
            args.type2_indices,
            args.type3_indices,
            logger=lambda msg: self.logger.info(f"{args.output} : {msg}"),
        )

        with open(args.output, "w", newline="") as f:
            writer = csv.writer(f)
            writer.writerow(["H", "b"])
            for H, b in zip(edge_heights, edge_widths):
                writer.writerow([H, b])

        self.logger.info(f"Saved {args.output}.")



"""Information projection to allowed classes."""

import cvxpy as cp
import numpy as np

__all__ = [
    "signed_iproj",
]


def signed_iproj(p, target_indices):
    """Signed information projection distance to allowed classes.

    Parameters
    ----------
    p : array-like, shape (K,)
        Probability distribution over K classes.
    target_indices : list of int
        List of target class indices to project onto.

    Returns
    -------
    signed_distance : float
        Signed information projection distance to the allowed classes.
    projected_distribution : array-like, shape (K,)
        Probability distribution after projection.

    Examples
    --------
    >>> import numpy as np
    >>> from heavyedge_features.iproj import signed_iproj
    >>> p = np.array([0.1, 0.7, 0.2])
    >>> target_indices = [0, 2]
    >>> dist, q = signed_iproj(p, target_indices)
    >>> dist
     np.float64(0.164...)
    >>> q
    array([0.117..., 0.441..., 0.441...])
    """
    p = np.maximum(p, 1e-12)  # Avoid log(0)
    p = p / p.sum()

    dists = []
    qs = []
    for target_index in target_indices:
        dist, q = _class_dist(p, target_index)
        dists.append(dist)
        qs.append(q)
    idx = np.argmin(dists)
    min_dist = dists[idx]
    mindist_q = qs[idx]

    if np.argmax(p) in target_indices:
        sign = -1  # signed distance
    else:
        sign = 1
    signed_distance, projected_distribution = sign * min_dist, mindist_q
    return signed_distance, projected_distribution


def _class_dist(p, i):
    K = len(p)

    dists = []
    qs = []
    for j in range(K):
        if j == i:
            continue
        dist, q = _class_dist_ij(p, i, j)
        dists.append(dist)
        qs.append(q)
    idx = np.argmin(dists)
    return dists[idx], qs[idx]


def _class_dist_ij(p, i, j):
    K = len(p)
    q = cp.Variable(K, nonneg=True)
    constraints = [cp.sum(q) == 1]

    # Constraints: q[i] == q[j], q[i] >= q[other]
    for k in range(K):
        if k == i:
            continue
        elif k == j:
            constraints.append(q[i] == q[k])
        else:
            constraints.append(q[i] >= q[k])
    objective = cp.Minimize(cp.sum(cp.kl_div(q, p)))
    return cp.Problem(objective, constraints).solve(), q.value
